local utf8 = require 'utf8'

local array = require 'array'
local my_utf8 = require 'my_utf8'
local t = require 'utils'
local rects = require 'rects'
local Loc = require 'loc'
local move = require 'move'
local keychord = require 'keychord'

local edit = {}

local I = {}  -- for file-local helpers, so I can refer to them before defining them
edit.internal = I
local _  -- idiom for unused variables

-- some constants people might like to tweak
Foreground_color = {r=0, g=0, b=0}  -- available to app
Cursor_color = {r=1, g=0, b=0}  -- available to app
local Highlight_color = {r=0.7, g=0.7, b=0.9, a=0.4}  -- selected text

local colorize = require 'colorize'  -- must come after Foreground_color is defined

Line_number_width = 3  -- in ems; available to app
local Line_number_color = {r=0.4, g=0.4, b=0.4}

local Font_height = require 'font_height'

function edit.is_this_love_version_supported()
  local supported_versions = {
    ['11.5']=true,
    ['11.4']=true,
    ['11.3']=true,
    ['11.2']=true,
    ['11.1']=true,
    ['11.0']=true,
    ['12.0']=true,
  }
  return supported_versions[edit.love_version()]
end

function edit.preferred_love_version()
  return '11.5'
end

function edit.love_version()
  local major_version, minor_version = love.getVersion()
  return major_version..'.'..minor_version
end

local Last_focus_time, Last_resize_time
Cursor_time = 0  -- available to app

-- initialize editor
-- this should happen only once regardless of how many editors you create
function edit.load()
  -- for hysteresis in a few places
  Last_focus_time = 0  -- https://love2d.org/forums/viewtopic.php?p=249700
  Last_resize_time = 0

  -- blinking cursor (I assume you never show the cursor in more than one editor)
  Cursor_time = 0
end

function edit.new_from_defaults(top, left, right, bottom)
  local font = love.graphics.newFont(Font_height)
  return edit.new(top, left, right, bottom, font, Font_height)
end

function edit.new_from_settings(settings, top, left, right, bottom)
  local font = love.graphics.newFont(settings.font_height)
  local editor = edit.new(top, left, right, bottom, font, settings.font_height)
  edit.load_file(editor, settings.filename)
  editor.screen_top = settings.screen_top
  editor.cursor = settings.cursor
  if not I.valid_loc(editor, editor.screen_top)
       or not I.valid_loc(editor, editor.cursor) then
    I.scroll_to_top(editor)
  end
  return editor
end

function edit.resize(editor, w, h, right, bottom)
  editor.right = right
  editor.width = editor.right-editor.left
  editor.bottom = bottom
  Last_resize_time = Current_time
end

function edit.settings(editor)
  return {
    font_height=editor.font_height,
    filename=editor.filename,
    screen_top=editor.screen_top, cursor=editor.cursor
  }
end

function edit.load_file(editor, filename)
  editor.filename = filename
  editor.lines = {}
  if love.filesystem.getInfo(filename) then
    for line in love.filesystem.lines(editor.filename) do
      table.insert(editor.lines, {data=line})
    end
  end
  if #editor.lines == 0 then
    table.insert(editor.lines, {data=''})
  end
  I.scroll_to_top(editor)  -- clear out stale locations
  colorize.all(editor)
end

function edit.load_array(editor, lines)
  editor.lines = {}
  assert(#lines > 0)
  for _, line in ipairs(lines) do
    table.insert(editor.lines, {data=line})
  end
  colorize.all(editor)
end

function edit.focus(in_focus)
  if in_focus then
    Last_focus_time = Current_time
  end
end

function edit.update_all(dt)
  Current_time = Current_time + dt
  -- some hysteresis while resizing
  if Current_time < Last_resize_time + 0.1 then
    return
  end
  --
  Cursor_time = Cursor_time + dt
end

function edit.maybe_autosave(editor, dt)
  if editor.next_save and editor.next_save < Current_time then
    save_to_disk(editor)
  end
end

function edit.final_save(editor)
  if editor.next_save then
    save_to_disk(editor)
    -- give some time for the OS to flush everything to disk
    love.timer.sleep(0.1)
  end
end

function edit.new(top, left, right, bottom, font, font_height)
  local result = {
    top = math.floor(top),
    left = math.floor(left),
    right = math.floor(right),
    bottom = math.floor(bottom),
    width = right-left,

    font = font,
    font_height = font_height,
    line_height = math.floor(font_height*1.3),

    -- The editor is for editing an array of lines.
    -- The array of lines can never be empty; there must be at least one line for positioning a cursor at.
    lines = {{data=''}},  -- array of strings

    -- We need to track a couple of _locations_:
    screen_top = {line=1, pos=1},  -- location at top of screen, to start drawing from
    cursor = {line=1, pos=1},  -- location where editing will occur

    selection = {},
    -- some extra editor to compute selection between mouse press and release
    old_cursor = nil,
    old_selection = nil,
    mousepress_shift = nil,

    filename = love.filesystem.getSourceBaseDirectory()..'/lines.txt',  -- '/' should work even on Windows
    next_save = nil,

    -- undo
    history = {},
    next_history = 1,

    -- search
    search_term = nil,
    search_backup = nil,  -- stuff to restore when cancelling search
  }
  return result
end

function edit.draw(editor, fg, hide_cursor, show_line_numbers)
  local cursor_or_mouse_loc = editor.cursor
  if editor.mouse_down then
    cursor_or_mouse_loc = I.loc_at_mouse(editor)
  end
  local y = editor.top
  for line_index = editor.screen_top.line, #editor.lines do
    local loc
    if line_index == editor.screen_top.line then
      loc = editor.screen_top
    else
      loc = {line=line_index, pos=1}
    end
    local rect = rects.compute(editor, loc, editor.bottom - y)
    if show_line_numbers and loc.pos == 1 then
      local w = Line_number_width * editor.font:getWidth('m')
      t.set_color_from_table(Line_number_color)
      love.graphics.print(loc.line, editor.left-w+10, y)
    end
    for _,s in ipairs(rect.screen_line_rects) do
--?       love.graphics.setColor(0.8, 0.8, 0.8)
--?       love.graphics.rectangle('fill', editor.left+s.x,y+s.y, s.dx,s.dy)
      for _,c in ipairs(s.char_rects) do
--?         love.graphics.setColor(0.8, 0.9, 0.8)
--?         love.graphics.rectangle('fill', editor.left+c.x,y+c.y, c.dx-2,c.dy-1)
        if c.show_cursor then  -- only the primary place where the cursor can be is available for highlight
          if I.in_selection(editor, line_index, c.pos, cursor_or_mouse_loc)
              or I.in_search(editor, line_index, c.pos)
          then
            t.set_color_from_table(Highlight_color)
            love.graphics.rectangle('fill', editor.left+c.x, y+c.y, c.dx,c.dy)
          end
        end
        if c.draw then
          for _,s in ipairs(c.draw) do I.draw_shape(s, editor.left, y, fg) end
        end
        if c.data and not c.conceal then
          if fg == nil then
            if c.pos then
              t.set_color_from_table(c.fg or editor.colors[line_index][c.pos])
            else
              t.set_color_from_table(c.fg)
            end
          else
            t.set_color_from_table(c.fg or fg)
          end
          love.graphics.print(c.data, editor.left+c.x, y+c.y)
        end
        if not hide_cursor and line_index == editor.cursor.line then
          if c.pos == editor.cursor.pos and c.show_cursor then
            I.draw_text_cursor(editor, editor.left+c.x, y+c.y, c.dy)
          end
        end
      end
    end
    y = y + rect.dy
    if y + editor.line_height > editor.bottom then
      break
    end
  end
  if editor.search_term then
    I.draw_search_bar(editor)
  end
end

function edit.mouse_press(editor, mx,my, mouse_button)
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if editor.search_term then return end
  editor.mouse_down = mouse_button

  -- prepare for a drag selecting text
  -- delicate dance between cursor/selection and old cursor/selection
  -- scenarios:
  --  regular press+release: sets cursor, clears selection
  --  shift press+release:
  --    sets selection to old cursor if not set otherwise leaves it untouched
  --    sets cursor
  --  press and hold to start a selection: sets selection on press, cursor on release
  --  press and hold, then press shift: ignore shift
  --    i.e. mouse_release should never look at shift editor
  editor.old_cursor = editor.cursor
  editor.old_selection = editor.selection
  editor.mousepress_shift = keychord.shift_down()
  editor.selection = move.to_loc(editor, mx,my)
end

function edit.mouse_release(editor, mx,my, mouse_button)
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if editor.search_term then return end
  editor.mouse_down = nil

  editor.cursor = move.to_loc(editor, mx,my)

  -- clean up
  if editor.mousepress_shift then
    if editor.old_selection.line == nil then
      editor.selection = editor.old_cursor
    else
      editor.selection = editor.old_selection
    end
  end
  editor.old_cursor, editor.old_selection, editor.mousepress_shift = nil
  if Loc.eq(editor.cursor, editor.selection) then
    editor.selection = {}
  end
end

function edit.mouse_wheel_move(editor, dx,dy)
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if dy > 0 then
    editor.cursor = t.deepcopy(editor.screen_top)
    for i=1,math.floor(dy) do
      move.up_arrow(editor)
    end
  elseif dy < 0 then
    editor.cursor = move.to_loc(editor, editor.left, editor.bottom-1)
    assert(editor.cursor)
    for i=1,math.floor(-dy) do
      move.down_arrow(editor)
    end
  end
end

function I.scroll_to_top(editor)
  assert(#editor.lines > 0)
  editor.screen_top = {line=1, pos=1}
  editor.cursor = {line=1, pos=1}
end

function I.valid_loc(editor, loc)
  assert(#editor.lines > 0)
  if loc == nil then return end
  if loc.line > #editor.lines then return end
  if loc.pos > #editor.lines[loc.line].data then return end
  return true
end

function I.schedule_save(editor)
  if editor.next_save == nil then
    editor.next_save = Current_time + 3  -- short enough that you're likely to still remember what you did
  end
end

I.buffer_modified_handlers = {}

function edit.signal_buffer_modified(editor)
  for _, handler in ipairs(I.buffer_modified_handlers) do
    handler(editor)
  end
end

table.insert(I.buffer_modified_handlers, I.schedule_save)
table.insert(I.buffer_modified_handlers, colorize.all)

---- keyboard handling

function edit.text_input(editor, t, readonly)
  -- ignore events for some time after window in focus (mostly alt-tab)
  if Current_time < Last_focus_time + 0.01 then
    return
  end
  --
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  if love.mouse.isDown(1) then return end
  if editor.search_term then
    editor.search_term = editor.search_term..t
    I.search_next(editor)
    return
  end
  if move.to_coord(editor, editor.cursor) == nil then return end  -- cursor is off screen
  if keychord.any_modifier_down() and key_down(t) then
    -- The modifiers didn't change the key. Handle it in keychord_press.
    return
  end
  if not readonly then
    local before = snapshot(editor, editor.cursor.line)
    -- Try to support mobile keyboard swipe/autocomplete/paste as best we
    -- can.
    --
    -- Mobile paste still works poorly. It's unreliable, often truncates
    -- large pastes, and inserts newlines twice, first keypress events for
    -- all characters in a paste and _then_ textinput for all characters.
    edit.insert_text(editor, t)
    record_undo_event(editor, {before=before, after=snapshot(editor, editor.cursor.line)})
    edit.signal_buffer_modified(editor)
  end
end

function edit.keychord_press(editor, chord, key, readonly)
  -- ignore events for some time after window in focus (mostly alt-tab)
  if Current_time < Last_focus_time + 0.01 then
    return
  end
  --
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
  local cursor_on_screen = move.to_coord(editor, editor.cursor)
  if not readonly and editor.selection.line and
      cursor_on_screen and
      -- printable character created using shift key => delete selection
      -- (we're not creating any ctrl-shift- or alt-shift- combinations using regular/printable keys)
      (not keychord.shift_down() or utf8.len(key) == 1) and
      chord ~= 'C-a' and chord ~= 'C-c' and chord ~= 'C-x' and chord ~= 'backspace' and chord ~= 'delete' and chord ~= 'C-z' and chord ~= 'C-y' and not keychord.is_cursor_movement(key) then
    I.delete_selection_and_record_undo_event(editor)
  end
  if editor.search_term then
    if chord == 'escape' then
      editor.search_term = nil
      editor.cursor = editor.search_backup.cursor
      editor.screen_top = editor.search_backup.screen_top
      editor.search_backup = nil
    elseif chord == 'return' then
      editor.search_term = nil
      editor.search_backup = nil
    elseif chord == 'backspace' then
      local len = utf8.len(editor.search_term)
      local byte_offset = my_utf8.offset(editor.search_term, len)
      editor.search_term = string.sub(editor.search_term, 1, byte_offset-1)
      editor.cursor = t.deepcopy(editor.search_backup.cursor)
      editor.screen_top = t.deepcopy(editor.search_backup.screen_top)
      I.search_next(editor)
    elseif chord == 'down' then
      if #editor.search_term > 0 then
        move.right_arrow(editor)
        I.search_next(editor)
      end
    elseif chord == 'up' then
      I.search_previous(editor)
    end
    return
  elseif chord == 'C-f' then
    editor.search_term = ''
    editor.search_backup = {
      cursor=t.deepcopy(editor.cursor),
      screen_top=t.deepcopy(editor.screen_top),
    }
  -- zoom
  elseif chord == 'C-=' then
    edit.update_font_settings(editor, editor.font_height+2)
  elseif chord == 'C--' then
    if editor.font_height > 2 then
      edit.update_font_settings(editor, editor.font_height-2)
    end
  elseif chord == 'C-0' then
    edit.update_font_settings(editor, 20)
  -- undo
  elseif not readonly and chord == 'C-z' then
    local event = I.undo_event(editor)
    if event then
      local src = event.before
      editor.screen_top = t.deepcopy(src.screen_top)
      editor.cursor = t.deepcopy(src.cursor)
      editor.selection = t.deepcopy(src.selection)
      I.patch(editor.lines, event.after, event.before)
      edit.signal_buffer_modified(editor)
    end
  elseif not readonly and chord == 'C-y' then
    local event = I.redo_event(editor)
    if event then
      local src = event.after
      editor.screen_top = t.deepcopy(src.screen_top)
      editor.cursor = t.deepcopy(src.cursor)
      editor.selection = t.deepcopy(src.selection)
      I.patch(editor.lines, event.before, event.after)
      edit.signal_buffer_modified(editor)
    end
  -- clipboard
  elseif chord == 'C-a' and cursor_on_screen then
    editor.selection = {line=1, pos=1}
    editor.cursor = {line=#editor.lines, pos=utf8.len(editor.lines[#editor.lines].data)+1}
  elseif chord == 'C-c' then
    local s = I.selection(editor)
    if s then
      love.system.setClipboardText(s)
    end
  elseif not readonly and chord == 'C-x' and cursor_on_screen then
    local s = I.cut_selection(editor, editor.left, editor.right)
    if s then
      love.system.setClipboardText(s)
    end
    edit.signal_buffer_modified(editor)
  elseif not readonly and chord == 'C-v' and cursor_on_screen then
    local before_line = editor.cursor.line
    local before = snapshot(editor, before_line)
    local clipboard_data = love.system.getClipboardText()
    edit.insert_text(editor, clipboard_data)
    record_undo_event(editor, {before=before, after=snapshot(editor, before_line, editor.cursor.line)})
    edit.signal_buffer_modified(editor)
  --== shortcuts that mutate text
  elseif not readonly and chord == 'return' and cursor_on_screen then
    local before_line = editor.cursor.line
    local before = snapshot(editor, before_line)
    I.insert_return_at_cursor(editor)
    move.maybe_snap_cursor_to_bottom_of_screen(editor)
    record_undo_event(editor, {before=before, after=snapshot(editor, before_line, editor.cursor.line)})
    edit.signal_buffer_modified(editor)
  elseif not readonly and chord == 'tab' and cursor_on_screen then
    local before = snapshot(editor, editor.cursor.line)
    I.insert_char_at_cursor(editor, '\t')
    move.maybe_snap_cursor_to_bottom_of_screen(editor)
    record_undo_event(editor, {before=before, after=snapshot(editor, editor.cursor.line)})
    edit.signal_buffer_modified(editor)
  elseif not readonly and chord == 'backspace' and cursor_on_screen then
    if editor.selection.line then
      I.delete_selection_and_record_undo_event(editor)
      edit.signal_buffer_modified(editor)
      return
    end
    local before
    if editor.cursor.pos > 1 then
      before = snapshot(editor, editor.cursor.line)
      local byte_start = utf8.offset(editor.lines[editor.cursor.line].data, editor.cursor.pos-1)
      local byte_end = utf8.offset(editor.lines[editor.cursor.line].data, editor.cursor.pos)
      if byte_start then
        if byte_end then
          editor.lines[editor.cursor.line].data = string.sub(editor.lines[editor.cursor.line].data, 1, byte_start-1)..string.sub(editor.lines[editor.cursor.line].data, byte_end)
        else
          editor.lines[editor.cursor.line].data = string.sub(editor.lines[editor.cursor.line].data, 1, byte_start-1)
        end
        editor.cursor.pos = editor.cursor.pos-1
      end
    elseif editor.cursor.line > 1 then
      before = snapshot(editor, editor.cursor.line-1, editor.cursor.line)
      -- join lines
      editor.cursor.pos = utf8.len(editor.lines[editor.cursor.line-1].data)+1
      editor.lines[editor.cursor.line-1].data = editor.lines[editor.cursor.line-1].data..editor.lines[editor.cursor.line].data
      table.remove(editor.lines, editor.cursor.line)
      editor.cursor.line = editor.cursor.line-1
    end
    if editor.screen_top.line > #editor.lines then
      editor.screen_top = vert(editor, editor.cursor, 0)
    elseif Loc.lt(editor.cursor, editor.screen_top) then
      move.maybe_snap_cursor_to_top_of_screen(editor)
    end
    assert(Loc.le(editor.screen_top, editor.cursor), ('screen_top (line=%d,pos=%d) is below cursor (line=%d,pos=%d)'):format(editor.screen_top.line, editor.screen_top.pos or -1, editor.cursor.line, editor.cursor.pos or -1))
    record_undo_event(editor, {before=before, after=snapshot(editor, editor.cursor.line)})
    edit.signal_buffer_modified(editor)
  elseif not readonly and chord == 'delete' and cursor_on_screen then
    -- cursor in text line
    if editor.selection.line then
      I.delete_selection_and_record_undo_event(editor)
      edit.signal_buffer_modified(editor)
      return
    end
    local before
    if editor.cursor.pos <= utf8.len(editor.lines[editor.cursor.line].data) then
      before = snapshot(editor, editor.cursor.line)
    else
      before = snapshot(editor, editor.cursor.line, editor.cursor.line+1)
    end
    if editor.cursor.pos <= utf8.len(editor.lines[editor.cursor.line].data) then
      local byte_start = utf8.offset(editor.lines[editor.cursor.line].data, editor.cursor.pos)
      local byte_end = utf8.offset(editor.lines[editor.cursor.line].data, editor.cursor.pos+1)
      if byte_start then
        if byte_end then
          editor.lines[editor.cursor.line].data = string.sub(editor.lines[editor.cursor.line].data, 1, byte_start-1)..string.sub(editor.lines[editor.cursor.line].data, byte_end)
        else
          editor.lines[editor.cursor.line].data = string.sub(editor.lines[editor.cursor.line].data, 1, byte_start-1)
        end
        -- no change to editor.cursor.pos
      end
    elseif editor.cursor.line < #editor.lines then
      -- join lines
      editor.lines[editor.cursor.line].data = editor.lines[editor.cursor.line].data..editor.lines[editor.cursor.line+1].data
      table.remove(editor.lines, editor.cursor.line+1)
    end
    record_undo_event(editor, {before=before, after=snapshot(editor, editor.cursor.line)})
    edit.signal_buffer_modified(editor)
  --== shortcuts that move the cursor
  elseif chord == 'left' and cursor_on_screen then
    move.left_arrow(editor)
    editor.selection = {}
  elseif chord == 'right' and cursor_on_screen then
    move.right_arrow(editor)
    editor.selection = {}
  elseif chord == 'S-left' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.left_arrow(editor)
  elseif chord == 'S-right' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.right_arrow(editor)
  -- C- hotkeys reserved for drawings, so we'll use M-
  elseif chord == 'M-left' and cursor_on_screen then
    move.word_left(editor)
    editor.selection = {}
  elseif chord == 'M-right' and cursor_on_screen then
    move.word_right(editor)
    editor.selection = {}
  elseif chord == 'M-S-left' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.word_left(editor)
  elseif chord == 'M-S-right' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.word_right(editor)
  elseif chord == 'home' and cursor_on_screen then
    move.start_of_line(editor)
    editor.selection = {}
  elseif chord == 'end' and cursor_on_screen then
    move.end_of_line(editor)
    editor.selection = {}
  elseif chord == 'S-home' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.start_of_line(editor)
  elseif chord == 'S-end' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.end_of_line(editor)
  elseif chord == 'up' and cursor_on_screen then
    move.up_arrow(editor)
    editor.selection = {}
  elseif chord == 'down' and cursor_on_screen then
    move.down_arrow(editor)
    editor.selection = {}
  elseif chord == 'S-up' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.up_arrow(editor)
  elseif chord == 'S-down' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.down_arrow(editor)
  elseif chord == 'pageup' and cursor_on_screen then
    move.pageup(editor)
    editor.selection = {}
  elseif chord == 'pagedown' and cursor_on_screen then
    move.pagedown(editor)
    editor.selection = {}
  elseif chord == 'S-pageup' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.pageup(editor)
  elseif chord == 'S-pagedown' and cursor_on_screen then
    if editor.selection.line == nil then
      editor.selection = t.deepcopy(editor.cursor)
    end
    move.pagedown(editor)
  end
end

function edit.insert_text(editor, s)
  for _,code in utf8.codes(s) do
    local c = utf8.char(code)
    if c == '\n' then
      I.insert_return_at_cursor(editor)
    else
      I.insert_char_at_cursor(editor, c)
    end
  end
  move.maybe_snap_cursor_to_bottom_of_screen(editor)
end

function edit.clear(editor)
  editor.lines = {{data=''}}
  editor.cursor = {line=1, pos=1}
  editor.screen_top = {line=1, pos=1}
  editor.selection = {}
  colorize.all(editor)
end

function edit.key_release(editor, key, scancode)
  -- ignore events for some time after window in focus (mostly alt-tab)
  if Current_time < Last_focus_time + 0.01 then
    return
  end
  --
  Cursor_time = 0  -- ensure cursor is visible immediately after it moves
end

function edit.update_font_settings(editor, font_height)
  editor.font_height = font_height
  editor.font = love.graphics.newFont(editor.font_height)
  editor.line_height = math.floor(font_height*1.3)
end

function I.draw_text_cursor(editor, x, y, dy)
  -- blink every 0.5s
  if math.floor(Cursor_time*2)%2 == 0 then
    t.set_color_from_table(Cursor_color)
    love.graphics.rectangle('fill', x,y, 3,dy)
  end
end

function I.insert_char_at_cursor(editor, t)
  local byte_offset = my_utf8.offset(editor.lines[editor.cursor.line].data, editor.cursor.pos)
  editor.lines[editor.cursor.line].data = string.sub(editor.lines[editor.cursor.line].data, 1, byte_offset-1)..t..string.sub(editor.lines[editor.cursor.line].data, byte_offset)
  editor.cursor.pos = editor.cursor.pos+1
end

function I.insert_return_at_cursor(editor)
  local byte_offset = my_utf8.offset(editor.lines[editor.cursor.line].data, editor.cursor.pos)
  table.insert(editor.lines, editor.cursor.line+1, {data=string.sub(editor.lines[editor.cursor.line].data, byte_offset)})
  editor.lines[editor.cursor.line].data = string.sub(editor.lines[editor.cursor.line].data, 1, byte_offset-1)
  editor.cursor = {line=editor.cursor.line+1, pos=1}
  move.maybe_snap_cursor_to_bottom_of_screen(editor)
end

-- mappings only to non-printable keys; leave out mappings that textinput will handle
local Numlock_off = {
  kp0='insert',
  kp1='end',
  kp2='down',
  kp3='pagedown',
  kp4='left',
  -- numpad 5 translates to nothing
  kp6='right',
  kp7='home',
  kp8='up',
  kp9='pageup',
  ['kp.']='delete',
  -- LÖVE handles keypad operators in textinput
  -- what's with the `kp=` and `kp,` keys? None of my keyboards have one.
  -- Hopefully LÖVE handles them as well in textinput.
  kpenter='enter',
  kpdel='delete',
}
local Numlock_on = {
  kpenter='enter',
  kpdel='delete',
}
function I.translate_numlock(key)
  if love.keyboard.isModifierActive('numlock') then
    return Numlock_on[key] or key
  else
    return Numlock_off[key] or key
  end
  return key
end

function I.draw_shape(shape, left, top, fg)
  t.set_color_from_table(shape.fg or fg)
  if shape.type == 'rect' then
    love.graphics.rectangle(shape.mode, left+shape.x, top+shape.y, shape.dx, shape.dy)
  elseif shape.type == 'line' then
    love.graphics.line(left+shape.x1, top+shape.y1, left+shape.x2, top+shape.y2)
  end
end

-- helpers for file operations

function save_to_disk(editor)  -- available to app
  if editor.filename == nil then
    editor.next_save = nil
    return
  end
  local lines = {}
  for _, line in ipairs(editor.lines) do
    table.insert(lines, line.data)
  end
  local contents = table.concat(lines, '\n')
  local success, message = love.filesystem.write(editor.filename, contents)
  if not success then
    error('failed to write to "'..editor.filename..'": '..message)
  end
  editor.next_save = nil
end

-- helpers for the search bar (C-f)

function I.draw_search_bar(editor)
  local screen_width, screen_height = love.window.getMode()
  local h = editor.line_height+2
  local y = screen_height-h
  love.graphics.setColor(0.9,0.9,0.9)
  love.graphics.rectangle('fill', 0, y-10, screen_width-1, h+8)
  love.graphics.setColor(0.6,0.6,0.6)
  love.graphics.line(0, y-10, screen_width-1, y-10)
  love.graphics.setColor(1,1,1)
  love.graphics.rectangle('fill', 20, y-6, screen_width-40, h+2, 2,2)
  love.graphics.setColor(0.6,0.6,0.6)
  love.graphics.rectangle('line', 20, y-6, screen_width-40, h+2, 2,2)
  t.set_color_from_table(Foreground_color)
  love.graphics.print(editor.search_term, 25,y-5)
  I.draw_text_cursor(editor, 25+editor.font:getWidth(editor.search_term),y-5, editor.line_height)
end

function I.search_next(editor)
  if #editor.search_term == 0 then return end
  local offset
  -- search current line from cursor
  local curr_pos = editor.cursor.pos
  local curr_line = editor.lines[editor.cursor.line].data
  local curr_offset = my_utf8.offset(curr_line, curr_pos)
  offset = I.find(curr_line, editor.search_term, curr_offset, --[[literal]] true)
  if offset then
    editor.cursor.pos = utf8.len(curr_line, 1, offset)
  end
  if offset == nil then
    -- search lines below cursor
    for i=editor.cursor.line+1,#editor.lines do
      local curr_line = editor.lines[i].data
      offset = I.find(curr_line, editor.search_term, --[[from start]] nil, --[[literal]] true)
      if offset then
        editor.cursor = {line=i, pos=utf8.len(curr_line, 1, offset)}
        break
      end
    end
  end
  if offset == nil then
    -- wrap around
    for i=1,editor.cursor.line-1 do
      local curr_line = editor.lines[i].data
      offset = I.find(curr_line, editor.search_term, --[[from start]] nil, --[[literal]] true)
      if offset then
        editor.cursor = {line=i, pos=utf8.len(curr_line, 1, offset)}
        break
      end
    end
  end
  if offset == nil then
    -- search current line until cursor
    local curr_line = editor.lines[editor.cursor.line].data
    offset = I.find(curr_line, editor.search_term, --[[from start]] nil, --[[literal]] true)
    local pos = utf8.len(curr_line, 1, offset)
    if pos and pos < editor.cursor.pos then
      editor.cursor.pos = pos
    end
  end
  if offset then
    move.maybe_snap_cursor_to_bottom_of_screen(editor)
  else
    -- roll back
    editor.cursor = t.deepcopy(editor.search_backup.cursor)
    editor.screen_top = t.deepcopy(editor.search_backup.screen_top)
  end
end

function I.search_previous(editor)
  if #editor.search_term == 0 then return end
  local offset
  -- search current line before cursor
  local curr_pos = editor.cursor.pos
  local curr_line = editor.lines[editor.cursor.line].data
  local curr_offset = my_utf8.offset(curr_line, curr_pos)
  offset = I.rfind(curr_line, editor.search_term, curr_offset-1, --[[literal]] true)
  if offset then
    editor.cursor.pos = utf8.len(curr_line, 1, offset)
  end
  if offset == nil then
    -- search lines above cursor
    for i=editor.cursor.line-1,1,-1 do
      local curr_line = editor.lines[i].data
      offset = I.rfind(curr_line, editor.search_term, --[[from end]] nil, --[[literal]] true)
      if offset then
        editor.cursor = {line=i, pos=utf8.len(curr_line, 1, offset)}
        break
      end
    end
  end
  if offset == nil then
    -- wrap around
    for i=#editor.lines,editor.cursor.line+1,-1 do
      local curr_line = editor.lines[i].data
      offset = I.rfind(curr_line, editor.search_term, --[[from end]] nil, --[[literal]] true)
      if offset then
        editor.cursor = {line=i, pos=utf8.len(curr_line, 1, offset)}
        break
      end
    end
  end
  if offset == nil then
    -- search current line after cursor
    local curr_line = editor.lines[editor.cursor.line].data
    offset = I.rfind(curr_line, editor.search_term, --[[from end]] nil, --[[literal]] true)
    local pos = utf8.len(curr_line, 1, offset)
    if pos and pos > editor.cursor.pos then
      editor.cursor.pos = pos
    end
  end
  if offset then
    move.maybe_snap_cursor_to_top_of_screen(editor)
  else
    -- roll back
    editor.cursor = t.deepcopy(editor.search_backup.cursor)
    editor.screen_top = t.deepcopy(editor.search_backup.screen_top)
  end
end

-- return true if cursor marks start of search term and line_index,pos is in that region
function I.in_search(editor, line_index, pos)
  if editor.search_term == nil then return false end
  if #editor.search_term == 0 then return false end
  if line_index ~= editor.cursor.line then return false end
  return I.find_at(editor.lines[line_index].data, editor.search_term, editor.cursor.pos)
    and editor.cursor.pos <= pos and pos <= editor.cursor.pos+utf8.len(editor.search_term)-1
end

function I.find_at(data, pat, pos)
  local offset = utf8.offset(data, pos)
  return data:sub(offset, offset+#pat-1) == pat
end

function I.find(s, pat, i, plain)
  if s == nil then return end
  return s:find(pat, i, plain)
end

-- TODO: avoid the expensive reverse() operations
-- Particularly if we only care about literal matches, we don't need all of string.find
function I.rfind(s, pat, i, plain)
  if s == nil then return end
  if #pat == 0 then return #s end
  local rs = s:reverse()
  local rpat = pat:reverse()
  if i == nil then i = #s end
  local ri = #s - i + 1
  local rendpos = rs:find(rpat, ri, plain)
  if rendpos == nil then return nil end
  local endpos = #s - rendpos + 1
  assert (endpos >= #pat, ('rfind: endpos %d should be >= #pat %d at this point'):format(endpos, #pat))
  return endpos-#pat+1
end

--? function test_rfind()
--?   check_eq(I.rfind('abc', ''), 3, 'empty pattern')
--?   check_eq(I.rfind('abc', 'c'), 3, 'final char')
--?   check_eq(I.rfind('acbc', 'c', 3), 2, 'previous char')
--?   check_nil(I.rfind('abc', 'd'), 'missing char')
--?   check_nil(I.rfind('abc', 'c', 2), 'no more char')
--? end

-- helpers for selecting portions of text

function I.loc_at_mouse(editor)
  local mx,my = love.mouse.getPosition()
  return move.to_loc(editor, mx, my)
end

function I.cut_selection(editor)
  if editor.selection.line == nil then return end
  local result = I.selection(editor)
  I.delete_selection_and_record_undo_event(editor)
  return result
end

function I.delete_selection_and_record_undo_event(editor)
  if editor.selection.line == nil then return end
  local minl,maxl = t.minmax(editor.selection.line, editor.cursor.line)
  local before = snapshot(editor, minl, maxl)
  I.delete_selection_without_undo(editor)
  record_undo_event(editor, {before=before, after=snapshot(editor, editor.cursor.line)})
end

function I.delete_selection_without_undo(editor)
  if editor.selection.line == nil then return end
  -- min,max = sorted(editor.selection,editor.cursor)
  local minl,minp = editor.selection.line,editor.selection.pos
  local maxl,maxp = editor.cursor.line,editor.cursor.pos
  if minl > maxl then
    minl,maxl = maxl,minl
    minp,maxp = maxp,minp
  elseif minl == maxl then
    if minp > maxp then
      minp,maxp = maxp,minp
    end
  end
  -- update editor.cursor and editor.selection
  editor.cursor.line = minl
  editor.cursor.pos = minp
  if Loc.lt(editor.cursor, editor.screen_top) then
    editor.screen_top = move.internal.loc_hor(editor, editor.cursor, editor.left)
  end
  editor.selection = {}
  -- delete everything between min (inclusive) and max (exclusive)
  local min_offset = my_utf8.offset(editor.lines[minl].data, minp)
  local max_offset = my_utf8.offset(editor.lines[maxl].data, maxp)
  if minl == maxl then
    editor.lines[minl].data = editor.lines[minl].data:sub(1, min_offset-1)..editor.lines[minl].data:sub(max_offset)
    return
  end
  assert(minl < maxl, ('minl %d not < maxl %d'):format(minl, maxl))
  local rhs = editor.lines[maxl].data:sub(max_offset)
  for i=maxl,minl+1,-1 do
    table.remove(editor.lines, i)
  end
  editor.lines[minl].data = editor.lines[minl].data:sub(1, min_offset-1)..rhs
end

function I.selection(editor)
  if editor.selection.line == nil then return end
  -- min,max = sorted(editor.selection,editor.cursor)
  local minl,minp = editor.selection.line,editor.selection.pos
  local maxl,maxp = editor.cursor.line,editor.cursor.pos
  if minl > maxl then
    minl,maxl = maxl,minl
    minp,maxp = maxp,minp
  elseif minl == maxl then
    if minp > maxp then
      minp,maxp = maxp,minp
    end
  end
  local min_offset = my_utf8.offset(editor.lines[minl].data, minp)
  local max_offset = my_utf8.offset(editor.lines[maxl].data, maxp)
  if minl == maxl then
    return editor.lines[minl].data:sub(min_offset, max_offset-1)
  end
  assert(minl < maxl, ('minl %d not < maxl %d'):format(minl, maxl))
  local result = {editor.lines[minl].data:sub(min_offset)}
  for i=minl+1,maxl-1 do
    table.insert(result, editor.lines[i].data)
  end
  table.insert(result, editor.lines[maxl].data:sub(1, max_offset-1))
  return table.concat(result, '\n')
end

function I.in_selection(editor, line_index, pos, cursor)
  if editor.selection.line == nil then return false end
  local curr = {line=line_index, pos=pos}
  if Loc.eq(cursor, editor.selection) then
    return false
  elseif Loc.lt(cursor, editor.selection) then
    return Loc.le(cursor, curr) and Loc.lt(curr, editor.selection)
  elseif Loc.lt(editor.selection, cursor) then
    return Loc.le(editor.selection, curr) and Loc.lt(curr, cursor)
  end
end

-- undo/redo by managing the sequence of events in the current session
-- based on https://github.com/akkartik/mu1/blob/master/edit/012-editor-undo.mu

-- makes a copy of lines on every single keystroke; will be inefficient with really long lines.
-- TODO: highlight stuff inserted by any undo/redo operation
-- TODO: coalesce multiple similar operations

function record_undo_event(editor, data)  -- available to app
  editor.history[editor.next_history] = data
  editor.next_history = editor.next_history+1
  for i=editor.next_history,#editor.history do
    editor.history[i] = nil
  end
end

function I.undo_event(editor)
  if editor.next_history > 1 then
--?     print('moving to history', editor.next_history-1)
    editor.next_history = editor.next_history-1
    local result = editor.history[editor.next_history]
    return result
  end
end

function I.redo_event(editor)
  if editor.next_history <= #editor.history then
--?     print('restoring history', editor.next_history+1)
    local result = editor.history[editor.next_history]
    editor.next_history = editor.next_history+1
    return result
  end
end

-- Copy all relevant global editor.
-- Make copies of objects; the rest of the app may mutate them in place, but undo requires immutable histories.
function snapshot(editor, s,e)  -- available to app
  -- Snapshot everything by default, but subset if requested.
  assert(s, 'failed to snapshot operation for undo history')
  if e == nil then
    e = s
  end
  assert(#editor.lines > 0, 'failed to snapshot operation for undo history')
  if s < 1 then s = 1 end
  if s > #editor.lines then s = #editor.lines end
  if e < 1 then e = 1 end
  if e > #editor.lines then e = #editor.lines end
  -- compare with edit.new
  local event = {
    screen_top=t.deepcopy(editor.screen_top),
    selection=t.deepcopy(editor.selection),
    cursor=t.deepcopy(editor.cursor),
    lines={},
    start_line=s,
    end_line=e,
    -- no filename; undo history is cleared when filename changes
  }
  for i=s,e do
    table.insert(event.lines, t.deepcopy(editor.lines[i]))
  end
  return event
end

function I.patch(lines, from, to)
--?   if #from.lines == 1 and #to.lines == 1 then
--?     assert(from.start_line == from.end_line)
--?     assert(to.start_line == to.end_line)
--?     assert(from.start_line == to.start_line)
--?     lines[from.start_line] = to.lines[1]
--?     return
--?   end
  assert(from.start_line == to.start_line, 'failed to patch undo operation')
  for i=from.end_line,from.start_line,-1 do
    table.remove(lines, i)
  end
  assert(#to.lines == to.end_line-to.start_line+1, 'failed to patch undo operation')
  for i=1,#to.lines do
    table.insert(lines, to.start_line+i-1, to.lines[i])
  end
end

return edit
